/* $Id: ClientTable.java,v 1.5 2006/12/21 09:12:04 bertl Exp $ */
package jsim;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.util.Date;
import java.sql.Time;
/**
The scrollable panel which contains a table. 
*/
class ClientTable extends JScrollPane {
   private final int nsegments;
   private TableModel dm;
   private DefaultTableColumnModel dcm;
   private String [][] data;
   /**  The table which will be displayed. */
   public JTable table;
   private final Color headerBackground;
   /** The renderer of the 0th column. */
  private final DefaultTableCellRenderer renderer0;
   /** The column model belonging to the table. */
   private final int NCOLS = 9;
   private int ncols;
   private int N;
    
  class Renderer0 extends DefaultTableCellRenderer {
    public Component getTableCellRendererComponent(JTable table,
       Object value, boolean isSelected, boolean hasFocus, int row,
       int column) { 
          Component c = renderer0.getTableCellRendererComponent(table, value, 
           isSelected, hasFocus, row, column);
       c.setBackground(headerBackground);
       return c;
       } 
   }
             
  class DataTableColumnModel extends DefaultTableColumnModel {
      class MyTableColumn extends TableColumn {
         MyTableColumn(int i, String name) {
            super(i);
            setHeaderValue(name);
            if (i==0) setCellRenderer(new Renderer0());
            setMinWidth(80);
            }       
         }
      
       DataTableColumnModel() {
         addColumn(new MyTableColumn(0,"segment"));
         addColumn(new MyTableColumn(1,"host"));
         addColumn(new MyTableColumn(2,"pid"));
         addColumn(new MyTableColumn(3,"explored"));
         addColumn(new MyTableColumn(4,"visited"));
         addColumn(new MyTableColumn(5,"enabled"));
         addColumn(new MyTableColumn(6,"transitions"));
         addColumn(new MyTableColumn(7,"deadlocks"));
         addColumn(new MyTableColumn(8,"ticks"));
         try {
          Runtime rt = Runtime.getRuntime();
          Process ps = rt.exec(new String[] {"/bin/sh","-c", "top -n 1 -b|grep PID"});
          BufferedReader in = new BufferedReader(new InputStreamReader(ps.getInputStream()));
          String s = in.readLine();
          StringTokenizer st = new StringTokenizer(s);
          ncols = st.countTokens();
          N = NCOLS + ncols;
          // System.out.println("Test: DataTableColumn:"+s);
          for (int i = NCOLS, k =0;st.hasMoreTokens();i++,k++) {
             MyTableColumn c = new MyTableColumn(i, st.nextToken());
             if (k>1 && k<11) addColumn(c);
             }    
         }
      catch (IOException e) {
         e.printStackTrace();
         System.exit(1);
         }
      }
      }
      
      public Vector update(Vector v) { 
               try {
	         Integer d = new Integer((String) (v.remove(0)));
                 int segment = d.intValue();
                 for (int i=3;i<NCOLS;i++) {
                    // System.err.println("Help:"+(String) v.get(i-1)+" "+i);
                    data[segment][i] = (String) v.remove(0);
                    }
                 ((AbstractTableModel) 
                 table.getModel()).fireTableRowsUpdated(segment, segment);
		 v.add(0, (Object) d);
		 return v;
                 }
            catch (NumberFormatException e) {
              e.printStackTrace();
              System.exit(1); 
            }
	    return null;
      }
      
      public void ended(int row) {
          // for (int i = NCOLS+2;i<N-2;i++)
              data[row][NCOLS+2]="stopped";
          ((AbstractTableModel) 
                 table.getModel()).fireTableRowsUpdated(row, row);
          }
          
      public void lastColumnsUpdate(int row, String s) {
          StringTokenizer st = new StringTokenizer(s);
          for (int i=NCOLS;st.hasMoreTokens() && i< N;i++)
                data[row][i]=st.nextToken();
          ((AbstractTableModel) 
                 table.getModel()).fireTableRowsUpdated(row, row);
          }
          
      public void firstUpdate(String hosts[], String pids[]) {
         for (int i=0;i<nsegments;i++) {
             data[i][1]=hosts[i];
             data[i][2]=pids[i];
             // System.out.println("Catched:"+hosts[i]+","+pids[i]);
             }
         ((AbstractTableModel) table.getModel()).fireTableDataChanged();
         }
            
      public ClientTable(final int nsegments) {
        this.nsegments = nsegments;
        dm = new AbstractTableModel() {
          public int getRowCount() {return nsegments;}
          public int getColumnCount() {return N;}
          public Object getValueAt(int r, int c) {
          return (Object) data[r][c];  
          }
        };
        DataTableColumnModel dcm =  new DataTableColumnModel();
        table = new JTable(dm, dcm);
        this.data = new String[nsegments][N];
        for (int i=0;i<nsegments;i++) 
          data[i][0]=String.valueOf(i);
        renderer0 = (DefaultTableCellRenderer) 
              (table.getTableHeader().getDefaultRenderer());
        headerBackground = table.getTableHeader().getBackground();
        table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        table.setCellSelectionEnabled(false);
        table.setColumnSelectionAllowed(false);
        table.setRowSelectionAllowed(false);
        JViewport vp = new JViewport();vp.setView(table);
        vp.setPreferredSize(table.getPreferredSize()); 
        setViewport(vp);
        Rectangle bound = new Rectangle(new Dimension(500,500)); 
        Rectangle port = new Rectangle(
                       new Dimension(vp.getPreferredSize()));
        bound = port.intersection(bound); 
        if (!bound.equals(port)) { 
            getViewport().setPreferredSize(bound.getSize());
            } 
        }
    }
