/* $Id: LogTable.java,v 1.1 2006/11/16 15:05:03 bertl Exp $ */
package jsim;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.util.Date;
import java.sql.Time;
import javax.swing.border.*;
/**
The scrollable panel which contains a table. 
*/
class LogTable extends JScrollPane {
   private AbstractTableModel dm;
   private DefaultTableColumnModel dcm;
   private ArrayList data = new ArrayList();
   /**  The table which will be displayed. */
   public JTable table;
   private final Color headerBackground;
   /** The renderer of the 0th column. */
   private final DefaultTableCellRenderer renderer0;
   /** The column model belonging to the table. */
   private final int NCOLS = 9;
    
  class Renderer0 extends DefaultTableCellRenderer {
    public Component getTableCellRendererComponent(JTable table,
       Object value, boolean isSelected, boolean hasFocus, int row,
       int column) { 
          Component c = renderer0.getTableCellRendererComponent(table, value, 
           isSelected, hasFocus, row, column);
       c.setBackground(headerBackground);
       return c;
       } 
   }
             
  class DataTableColumnModel extends DefaultTableColumnModel {
      class MyTableColumn extends TableColumn {
         MyTableColumn(int i, String name) {
            super(i);
            setHeaderValue(name);
            if (i==0) setCellRenderer(new Renderer0());
            setMinWidth(80);
            }       
         }
      
       DataTableColumnModel() {
         addColumn(new MyTableColumn(0,"level"));
         addColumn(new MyTableColumn(1,"explored"));
         addColumn(new MyTableColumn(2,"visited"));
         addColumn(new MyTableColumn(3,"enabled"));
         addColumn(new MyTableColumn(4,"transitions"));
         addColumn(new MyTableColumn(5,"tick_step"));
         addColumn(new MyTableColumn(6,"deadlocks"));
         addColumn(new MyTableColumn(7,"ticks"));
         addColumn(new MyTableColumn(8,"time"));
         }
      }
      
      public void update(Vector v) { 
               try {
                 int line = new Integer((String) (v.get(0))).intValue();
 // System.err.println("aap:"+line+"noot"+v.size()+"mies"+v.get(1));
                 if (line<data.size()) {
                      data.set(line, v.clone());
                      dm.fireTableRowsUpdated(line, line);
                      }
                 else {
                     data.add(v.clone());
                     dm.fireTableRowsInserted(line, line);
                     }
                 }
            catch (NumberFormatException e) {
              e.printStackTrace();
              System.exit(1); 
            }
      }
      
      String convertTime(String time) {
         try {
            int timeInSeconds = new Integer(time).intValue();
            int hours = timeInSeconds / 3600;
            timeInSeconds = timeInSeconds - (hours * 3600);
            int minutes = timeInSeconds / 60;
            timeInSeconds = timeInSeconds - (minutes * 60);
            return ""+hours+"h:"+minutes+"m:"+timeInSeconds+"s";
         }
         catch (NumberFormatException e) {
            e.printStackTrace();
            System.exit(1); 
            }
         return null;
         }
          
      public LogTable() {
        dm = new AbstractTableModel() {
          public int getRowCount() {return data.size()<10?10:data.size();}
          public int getColumnCount() {return NCOLS;}
          public Object getValueAt(int r, int c) {
          String s = r<data.size()?(String) (((Vector) data.get(r)).get(c+1)):"";
          if (c==8 && r<data.size()) s = convertTime(s);
          return (Object) s; 
          }
        };
        
        DataTableColumnModel dcm =  new DataTableColumnModel();
        table = new JTable(dm, dcm);
        renderer0 = (DefaultTableCellRenderer) 
              (table.getTableHeader().getDefaultRenderer());
        headerBackground = table.getTableHeader().getBackground();
        table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        table.setCellSelectionEnabled(false);
        table.setColumnSelectionAllowed(false);
        table.setRowSelectionAllowed(false);
        setViewportView(table);
        getViewport().setPreferredSize(table.getPreferredSize());
        setBorder(BorderFactory.createTitledBorder(
         BorderFactory.createEtchedBorder(), 
           "History",
          TitledBorder.CENTER, TitledBorder.ABOVE_TOP)); 
        }
    }
