/*$Id: MsgHandler.java,v 1.6 2005/11/10 16:11:47 bertl Exp $*/
package jsim;

import java.util.*;
import java.lang.*;
import javax.swing.*;
import java.io.*;
/**
 * Source code from "Java Distributed Computing", by Jim Farley.
 *
 * Class: Msg
 * Example: 6-2
 * Description: A base class for agents that utilize message-passing.
 */

class ListButton  {
   public String name, pos, marked, color;
        ListButton(String name, String pos, String marked, String color) {
           this.name = name;
           this.pos = pos;
           this.marked = marked;
           this.color = color;
        }
     }
     
class SummandButton  {
   public String pos, marked, smd;
        SummandButton(String pos, String marked, String smd) {
           this.pos = pos;
           this.marked = marked;
           this.smd = smd;
        }
     }
          
abstract class Handler implements Runnable
{
  // Static message handler for applications where only one is used
  // and needs to be globally accessible.
  // public static Handler current = null;
  public Handler current = null;
  DataInputStream msgIn;
  OutputStream msgOut;
  StreamTokenizer tokenizer;
  String msgEndToken = "#", msgButton = "#button",
  msgSummand = "#summand";

  public Handler(InputStream in, OutputStream out) {
    setStreams(in, out);
    current = this;
  }

  public Handler(InputStream in, OutputStream out, String endToken) {
    msgEndToken = endToken;
    setStreams(in, out);
    current = this;
  }

  protected void setStreams(InputStream in, OutputStream out) {
    msgIn = new DataInputStream(in);
    msgOut = out;
  }

  public Msg readMsg() throws IOException {
    DataInputStream din = msgIn;
    // if (din.available()==0) return null;
    String token = din.readUTF();
    // System.err.println("token:"+token);
    Msg msg = buildMessage((String) token);
    if (msg != null) {
      synchronized(msg.argList) {
      msg.clearArg();
      boolean msgEnd = false;
      while (!msgEnd) {
        token = din.readUTF();
        // System.err.println("QQ:"+token+":"+msgSummand);
        if (token.compareTo(msgEndToken) == 0)
          msgEnd = true;
        else
        if (token.compareTo(msgButton) == 0) {
          String lab = din.readUTF();
          String cmd = din.readUTF();
          String marked = din.readUTF();
          String color = din.readUTF();
          msg.addArg(new ListButton(lab, cmd, marked, color));
          }
        else
        if (token.compareTo(msgSummand) == 0) {
          String pos = din.readUTF();
          String marked = din.readUTF();
          String smd = din.readUTF();
          msg.addArg(new SummandButton(pos, marked, smd));
          }
        else {
          msg.addArg(token);
        }
        }
      }
    }
    return msg;
  }

/*  
  public void writeMsg(Msg msg) {
    DataOutputStream dout = new DataOutputStream(msgOut);
    try {
        // System.err.println("sendMsg:"+msg.getId());
        // dout.writeByte(255);
        dout.writeUTF(msg.getId());
        Vector args = msg.argList();
        int acnt = args.size();
        for (int i = 0; i < acnt; i++) {
          dout.writeUTF((String)args.elementAt(i));
        }
        dout.writeUTF(msgEndToken);
        dout.flush();
        // System.err.println("end sendMsg:"+msg.getId());
    }
    catch (IOException e) {
              // System.err.println(e);
              // e.printStackTrace();
     }
    }
*/
  
  public boolean sendMsg(Msg msg) {
    DataOutputStream dout = new DataOutputStream(msgOut);
    try {
        // System.err.println("sendMsg:"+msg.getId());
        dout.writeUTF(msg.getId());
        Vector args = msg.argList();
        int acnt = args.size();
        for (int i = 0; i < acnt; i++) {
          dout.writeUTF((String)args.elementAt(i));
        }
        dout.writeUTF(msgEndToken);
        dout.flush();
        return true;
        // System.err.println("end sendMsg:"+msg.getId());
    }
    catch (IOException e) {return false;}
  }

  public void run() {
    try {
      // System.err.println("Stat thread");
      while (true) {
        Msg msg = readMsg();
        if (msg != null)
          msg.Do();
      }
    }
    // Treat an IOException as a termination of the message
    // exchange, and let this message-processing thread die.
    catch (IOException e) {
       // System.err.println("stopped");
       }
  }

  protected abstract Msg buildMessage(String msgId);
}

public class MsgHandler extends Handler {
  Nsim nsim;
  Screen screen;
  Hashtable msg = new Hashtable();
  
  MsgHandler(InputStream in, OutputStream out, Nsim nsim,
     Object table[][]) {
     super(in, out);
     this.nsim = nsim;
     for (int i=0;i<table.length;i++) {
         // System.out.println("aap:"+table[i][0]+":"+table[i][1]);
         msg.put(table[i][0], table[i][1]);
         }
     }
     
   MsgHandler(InputStream in, OutputStream out, Screen screen,
     Object table[][]) {
     super(in, out);
     this.screen = screen;
     for (int i=0;i<table.length;i++) {
         // System.out.println("aap:"+table[i][0]+":"+table[i][1]);
         msg.put(table[i][0], table[i][1]);
         }
     }
  protected Msg buildMessage(String msgId) {
     // System.out.println("Got message" +msgId);
     Msg result = (Msg)(msg.get((Object) msgId));
     result.clearArg();
     // System.err.println("result="+result);
     return result;
     }
  }
