/* $Id: React.java,v 1.5 2003/04/16 11:47:00 bertl Exp $ */
package stateview;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import att.grappa.*;
import aterm.*;
import aterm.pure.*;
  

/**
* The class which defines the reactions on mouse clicks.
* Mouse clicks are meaningful on:
* <ul>
* <li> Edge
* <li> Remaining subgraph
* </ul>
* @author Bert Lisser
*/    
public class React extends GrappaAdapter {
/** The cluster belonging to this object */
  public final Cluster cluster;
/** The set of popup menus; each edge has one popup menu. */
  private JPopupMenu f[];
/** The table which will be visible after clicking in the state graph. */
  JDialog dialog;
/** The option pane containing tables about parameters */
  ShowClusterTables showclustertables;
  
/**
* Class constructor.
* @param cluster belongs to this object
*/         
  React(Cluster cluster) {
       this.cluster = cluster;
       f = new JPopupMenu[cluster.edges.length];
       }
/** clears the highlights of the edges belonging to a summand.
* @param d number of the summand 
*/  
      
     static private void clearHighlight(int d) {
      Iterator lines =  Stateview.lines[d].iterator(); 
      while (lines.hasNext()) {
        Bundle b = (Bundle) (lines.next());
        Element a = b.getLine();
        // System.err.println("setHighlight2 "+a);
        GrappaSupport.setHighlight(a, 1, GrappaConstants.HIGHLIGHT_OFF);
        a.getGraph().repaint();
        } 
      }
      
   static void clearHighlight() {
      for (int i=0;i<Data.nrows;i++) {;
        Iterator lines =  Stateview.lines[i].iterator(); 
        while (lines.hasNext()) {
        Bundle b = (Bundle) (lines.next());
        Edge a = (Edge) b.getLine();
        // System.err.println("setHighlight2 "+a);
        GrappaSupport.setHighlight(a, 1, GrappaConstants.HIGHLIGHT_OFF);
        GrappaSupport.setHighlight(a.getTail(), 1, GrappaConstants.HIGHLIGHT_OFF);
        GrappaSupport.setHighlight(a.getHead(), 1, GrappaConstants.HIGHLIGHT_OFF);
        a.getGraph().repaint();
        } 
      }
    }
      
/** highlights the edges belonging to a summand.
@param d number of the summand 
*/  
     private void setHighlight(int d, Node from, Node to) {
      Iterator lines =  Stateview.lines[d].iterator(); 
      while (lines.hasNext()) {
        Bundle b = (Bundle) (lines.next());
        Edge c = (Edge) b.getLine();
        // System.err.println("setHighlight2 "+a);
        if (from.getGraph()!= c.getGraph() || 
           (from==c.getTail()&&to==c.getHead())
           ) {
          GrappaSupport.setHighlight(c, 1, GrappaConstants.HIGHLIGHT_ON);
          c.getGraph().repaint();
          }
        }
      }
      
/** clear highlights of the edges belonging to a summand.
@param d number of the summand 
*/  
     private void clearHighlight(int d, Node from, Node to) {
      Iterator lines =  Stateview.lines[d].iterator(); 
      while (lines.hasNext()) {
        Bundle b = (Bundle) (lines.next());
        Edge c = (Edge) b.getLine();
        // System.err.println("setHighlight2 "+a);
        if (from.getGraph()!= c.getGraph() || 
           (from==c.getTail()&&to==c.getHead())
           ) {
          GrappaSupport.setHighlight(c, 1, GrappaConstants.HIGHLIGHT_OFF);
          c.getGraph().repaint();
          }
        }
      }
           
/** shows the table containing condition, action, and relevant next states
    belonging to a summand.
@param d number of the summand 
*/          
     public void showSelectedTable(int d) {
      ShowTable showTable = new ShowTable(d, this.cluster.parameter.name);
      showTable.setVisible(true);
      }
                     
  public void grappaClicked(Subgraph subg, final Element elem, GrappaPoint pt, int modifiers, 
     int clickCount, GrappaPanel panel) {
     if (elem==null) return;
     int mask = InputEvent.BUTTON1_MASK;
     if (elem.getType()==GrappaConstants.SUBGRAPH && (modifiers & mask) == mask) {
       if (dialog==null) { 
          showclustertables = 
          new ShowClusterTables(cluster); 
          dialog = showclustertables.makeDialog(this.cluster.parameter.name);
          if (dialog==null) return;
          showclustertables.MarkCells();
          }
       dialog.setVisible(true);
       }
     }
 
/** This class highlights the belonging edges if the mouse enters a menu entry,
* which belongs to a summand */
  
  class MyMouseListener implements MouseListener {
      private JMenuItem item;
      private Edge edge;
      private int last = -1;
      public void mouseClicked(MouseEvent e) {}
      public void mouseEntered(MouseEvent e) {
          int v = Integer.parseInt(item.getActionCommand());
          setHighlight(v, edge.getTail(), edge.getHead());
          last = v;
          }
      public void mouseExited(MouseEvent e) {
          clearHighlight(last);
          }
      public void mousePressed(MouseEvent e) {}
      public void mouseReleased(MouseEvent e) {}
      MyMouseListener(JMenuItem item, Edge edge) {
          this.item = item;
          this.edge = edge;
          }
      void offHighlight() {clearHighlight(last);}
      }
/** resets the edge colors to the original coloring */
   private void resetColor() {
      Bundle[] edges = this.cluster.edges;
      for (int i=0;i<edges.length;i++) {
         edges[i].clearColor();
         }
      for (int i=0;i<edges.length;i++) {
         edges[i].setColor();
         // System.out.println("Setcolor:"+edges[i].label);
         }
      this.cluster.stategraph.repaint();
      }
            
 /** displays a menu if the mouse enters an edge and clears that menus if the mouse 
 *   exits that edge */
 
   public String grappaTip(Subgraph subg, final Element elem, GrappaPoint pt, int modifiers, 
        GrappaPanel panel) {
        if (elem==null) return null;
        clearHighlight();
        if (elem.getType()==GrappaConstants.SUBGRAPH) {
        // return dialog != null 
        // && dialog.isVisible()? null: "Click for process information";
        for (int i=0;i<f.length;i++)
        if (f[i]!=null) f[i].setVisible(false);
        this.resetColor();
        return null;
        }
        /*
        if (elem.getType()==GrappaConstants.NODE) {
        // return dialog != null 
        // && dialog.isVisible()? null: "Click for process information";
        for (int i=0;i<f.length;i++)
        if (f[i]!=null) f[i].setVisible(false);
        if (this.cluster.selectedPar != null) {
             // System.err.println("Hallo "+this.cluster.selectedPar.name);
             InitialState root = new InitialState((Node) elem, 
             this.cluster.selectedPar);
             root.occursInSubgraph();
             }
        return null;
        }
        */
        if (elem.getType()==GrappaConstants.EDGE) {
           int d = Integer.parseInt(
              elem.getAttribute(GrappaConstants.TAG_ATTR).getStringValue());
           Bundle[] edges= cluster.edges;
           if (edges.length!=0 && showclustertables != null) {
                showclustertables.indexTable.clearSelection();
                Transition[] transitions = edges[d].transitions;
                  for (int j=0;j<transitions.length;j++) {
                     Transition transition = transitions[j];
                     Integer value = 
                        (Integer) showclustertables.transitions.get((Object) transition);
                   showclustertables.indexTable.addRowSelectionInterval(
                        value.intValue(), value.intValue());
                   }
                // clustertable.headerColumn.selectAll();   
               }
         this.resetColor();
         if (f[d]!=null) {
             if (f[d].isShowing()) {f[d].setVisible(true); return null;}
             else {
                 java.awt.geom.Point2D 
                 qt = panel.getTransform().transform((java.awt.geom.Point2D) pt, null);
                 f[d].show(panel, (int) qt.getX(),(int) qt.getY());
                 }
             }
         else {
             Transition[] transitions = edges[d].transitions;
             JPopupMenu popup = new JPopupMenu(this.cluster.parameter.name);
             // popup.add(this.cluster.parameter.name);
             // popup.addSeparator();
             for (int j = 0;j<transitions.length;j++) {
                   Transition transition = transitions[j];
                   JMenuItem item = popup.add(new AbstractAction(Data.summands[transition.sum].action
                      +"{"+ (transition.sum+1)+"}") {
                   public void actionPerformed(ActionEvent e) {
                      try {
                        // System.err.println(e.getActionCommand());
                        showSelectedTable(Integer.parseInt(e.getActionCommand()));
                        JMenuItem item = (JMenuItem) e.getSource();
                        Action action = item.getAction();
                        MyMouseListener mouse = (MyMouseListener) 
                              action.getValue("mouse");
                        mouse.offHighlight();
                        }
                      catch (NumberFormatException m) {
                           System.err.println(m.getMessage());
                           System.exit(1);
                        }
                      }
                   });
                   item.setActionCommand(String.valueOf(transition.sum));
                   Action action = item.getAction();
                   MyMouseListener mouse= new MyMouseListener(item, (Edge) elem); 
                   item.addMouseListener(mouse);
                   action.putValue("mouse",(Object) mouse);
                }
             popup.pack();     
             f[d] = popup;
             java.awt.geom.Point2D 
                  qt = panel.getTransform().transform((java.awt.geom.Point2D) pt, null);
             f[d].show(panel, (int) qt.getX(), (int) qt.getY()); 
             }    
           /* return dialog != null && dialog.isVisible()? null:
           "Click for information about actions and summands"; */
           return null;
           } 
        return null;      
        }
 }
// System.err.println("NODE:"+ d);
       //JOptionPane pane = new JOptionPane(table,JOptionPane.PLAIN_MESSAGE,
       //JOptionPane.DEFAULT_OPTION, (Icon) null, new String[]{"Dismiss"});
       // JDialog diag = pane.createDialog(null, elem.getName());
       // diag.setModal(false);
       // diag.setVisible(true);
