/* $Id: Stateview.java,v 1.6 2003/04/17 14:12:44 bertl Exp $ */
package stateview;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.util.jar.*;
import java.util.zip.ZipEntry;
import javax.swing.*;
import javax.swing.event.*;
import att.grappa.*;
import aterm.*;
import aterm.pure.*;
import javax.swing.text.EditorKit;
import javax.swing.text.html.*;

/** Reads commands from simulator. */
   
class ExecuteCommands extends Thread {
    private  DataInputStream input;
    private  Node[] node;
    private Hashtable lastEdge = new Hashtable(100);
    ExecuteCommands(InputStream input) {
        this.input = new DataInputStream(input);
        System.out.println("Start thread");
        this.node = new Node[Data.clusters.length];
        }
        
    synchronized public void clear() {
        for (int i=0;i<Data.clusters.length;i++) {
           StateGraph stateGraph = Data.clusters[i].stategraph;
           if (stateGraph!=null) {
               Graph graph = stateGraph.graph;
               GraphEnumeration es = graph.elements(GrappaConstants.NODE|GrappaConstants.EDGE);
               while (es.hasMoreElements()) {
                   Element e = es.nextGraphElement();
                   GrappaSupport.setHighlight(e, 1, GrappaConstants.HIGHLIGHT_OFF);
                   }
               graph.repaint();
               if (stateGraph.react != null) {
                   React react = stateGraph.react;
                   if (react.showclustertables!=null) {
                       react.showclustertables.indexTable.clearSelection();
                       }
                   }
               }
           }
        }
              
    synchronized private void getState() {
        try {
           short n = input.readShort();
           String[] s = new String[n];
           for (int i=0;i< n;i++) {
               s[i]= input.readUTF();
               // System.out.println("Ja:"+s[i]);
               }
           /*
           for (int i=0;i<Data.clusters.length;i++) {
               if (node[i]!=null) { 
                 GrappaSupport.setHighlight((Element) node[i], 1, GrappaConstants.HIGHLIGHT_OFF);
                 Edge edge = (Edge) lastEdge.get(node[i]);
                 if (edge!=null) 
                     GrappaSupport.setHighlight(edge, 1, GrappaConstants.HIGHLIGHT_OFF);
                 node[i].getGraph().repaint();
                 }
               }
           */
           clear();
           for (int i=0;i<Data.clusters.length;i++) {
              Cluster cluster = Data.clusters[i];
              // System.out.println("QQ pos:"+cluster.parameter.pos+" "+s[cluster.seq]);
              node[i] = (Node) 
                     (cluster.table.get((Object) s[cluster.parameter.pos]));
              if (node[i]!=null) {
                 GrappaSupport.setHighlight(node[i], 1, GrappaConstants.HIGHLIGHT_ON);
                 Edge edge = (Edge) lastEdge.get(node[i]);
                 if (edge!=null) 
                     GrappaSupport.setHighlight(edge, 1, GrappaConstants.HIGHLIGHT_ON);
                 node[i].getGraph().repaint();
                 }
              }  
           } catch(IOException fnf) {
           System.err.println(fnf.toString());
           System.exit(1);
           }  
        }
        
    synchronized private void getEdge() {
        try {
           short n = input.readShort();
           if (n<0) return;
           Iterator lines =  Stateview.lines[n].iterator(); 
           for (int i=0;lines.hasNext();i++) {
               Bundle b = (Bundle) (lines.next());
               Edge a = b.getLine();
               if (node[b.parent.id]==a.getTail()) {
                   lastEdge.put(a.getHead(), a);
               }
           }
        } catch(IOException fnf) {
           System.err.println(fnf.toString());
           System.exit(1);
           }  
        }
                      
    public void run() {
       String s = null;
       try {
           System.out.println("Start run");
           while ((s = input.readUTF())!=null) {
                // System.out.println("Ja:"+s);
                if (s.equals("alert")) {continue;}
                if (s.equals("state")) {getState();continue;}
                if (s.equals("edge")) {getEdge();continue;}
                if (s.equals("reset")) {clear();lastEdge.clear();continue;}
                if (s.equals("finish")) System.exit(0);
                System.out.println("Illegal tag:"+s);System.exit(1);
                }
         } catch(IOException fnf) {
	   // System.err.println(fnf.toString()+":"+((s!=null)?s:"EOF"));
	   return;
         }
    }
   }
   
/**
* Stateview is the main class. It gives access to all grappa information.
* @author Bert Lisser
*/
        
public class Stateview implements GrappaConstants
{  
   /** input stream from which the data is read */
   private static InputStream input;
   /** file from which the data is read */
   private static File file;
   /** array of sets of Grappa edges indexed by summand number */
   public static java.util.List[] lines;
   public static ExecuteCommands thread;
   public static Font font;
   
   static void ComputeBounds(GrappaPanel gp, Graph g) {
       GrappaBox bbox = new GrappaBox(g.getBoundingBox());
       if(bbox == null) return;
       GrappaSize margins = (GrappaSize)(g.getAttributeValue(MARGIN_ATTR));
       if(margins != null) {
            double x_margin = PointsPerInch * margins.width;
            double y_margin = PointsPerInch * margins.height;

            bbox.x -= x_margin;
            bbox.y -= y_margin;
            bbox.width += 2.0 * x_margin;
            bbox.height += 2.0 * y_margin;
        }
        gp.setSize(new Dimension((int)Math.ceil(bbox.getWidth()),(int)Math.ceil(bbox.getHeight()+22)));
        gp.setPreferredSize(new Dimension((int)Math.ceil(bbox.getWidth()),
            (int)Math.ceil(bbox.getHeight()))); 
   }
   
   static void DisplayHelp(String fileName) {
        final JEditorPane jep = new JEditorPane();
        // jep.setFont(font);
        jep.setEditable(false);
        EditorKit htmlKit = jep.getEditorKitForContentType("text/html");
        HTMLDocument doc = (HTMLDocument) htmlKit.createDefaultDocument();
        InputStream in = null;
        jep.setEditorKit(htmlKit);
        try {
            File file = new File(fileName);
            in = (InputStream) new FileInputStream(file);
            } 
        catch(FileNotFoundException fnf) {
            try {
               String classPath = System.getProperty("java.class.path",".");
               StringTokenizer st = new StringTokenizer(classPath, ":", false);
               while (st.hasMoreTokens()) {
                 String s = st.nextToken();
                 File f = new File(s);
                 // System.err.println("aap:"+f.getName());
                 if (f.isFile() && f.getName().equals("stateview.jar")) {
                 JarFile file = new JarFile(f);
                 ZipEntry zipEntry = file.getEntry("stateview/"+fileName);
                 if (zipEntry != null)
                    in = (InputStream) file.getInputStream(zipEntry);
                    break;
                 }
                }
              } 
            catch(FileNotFoundException e) {
                System.err.println(e);
                System.exit(1);
                }
            catch (IOException e) {
                System.err.println(e);
                System.exit(1);
                } 
            }
        if (in!=null) {
           try {
              jep.read(in,doc);
              jep.setCaretPosition(0);
              }
           catch (IOException e) {
              System.err.println(e);
              System.exit(1);
              }
        }   
        JScrollPane scrollPane = new JScrollPane(jep);
        // scrollPane.getViewport().updateUI();
        JFrame f = new JFrame("help info");
        // f.setFont(font);
        f.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        f.setContentPane(scrollPane);
        f.setSize(512,342);
        /*
        f.addWindowListener(new WindowAdapter() {
		    public void windowActivated(WindowEvent w) {
                    jep.scrollRectToVisible(new Rectangle(0,0));
		    }
		});
        */	
        f.show();
        }
            
   public static void main(String[] args) {
        // System.out.println("Development");
	input = System.in;
	if(args.length > 1) {
	    System.err.println("USAGE: java Stateview [input_graph_file]");
	    System.exit(1);
	} else if(args.length == 1) {
	    if(args[0].equals("-")) {
		input = null;
	    } else {
		try {
                    file = new File(args[0]);
		    input = (InputStream) new FileInputStream(file);
		} catch(FileNotFoundException fnf) {
		    System.err.println(fnf.toString());
		    System.exit(1);
		}
	    }
	}
        /*
        Font f[] = GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts();
        for (int i=0;i<f.length;i++) 
        {
        font = f[i];
        System.out.println("font:"+font);
        }
        */
        // font=font.deriveFont(14.0f);
        // font = new Font("dialoginput.italic",Font.PLAIN, 12);
        // font = new Font(null,Font.PLAIN, 12);
        ToolTipManager tooltipmanager = ToolTipManager.sharedInstance();
        tooltipmanager.setInitialDelay(1);
        Data.makeData(input);
        if (!file.delete()) 
             System.out.println(
             "Not possible to remove "+file.getAbsolutePath());
        lines = new java.util.List[Data.nrows];
        for (int i=0;i<Data.nrows;i++)
           lines[i] =  Collections.synchronizedList(new LinkedList());
        ClusterGraph cg = new ClusterGraph(Data.clustergraph, 
                 Data.clusters.length);
        cg.setVisible(true);
        // System.err.println("After visible"+ cg.getViewport().getPreferredSize());
        thread = new ExecuteCommands(System.in);
        thread.start();
       }

    }
    
